/* Pxys - An IRC ProxyScan, m_scanner.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * Scanner node.
 *
 * $Id: m_scanner.c,v 1.28 2003/06/26 13:43:01 mbuna Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "conf.h"
#include "debug.h"

#include "typedefs.h"

#include "m_array.h"
#include "m_cache.h"
#include "m_log.h"
#include "m_noscan.h"
#include "m_scanner.h"
#include "m_share.h"
#include "m_stats.h"

#include "ms_globals.h"
#include "ms_msgout.h"       /* to inform the console channel */
#include "ms_netutils.h"

#include "mp_listener.h"
#include "mp_negativity.h"
#include "mp_proxy.h"
#include "mp_socks.h"
#include "mp_wingate.h"

#include "xmalloc.h"

/* We use an Hoare's monitor to start the scanner:
 * - All scan threads are launched and wait on a condition
 *   (see scanner_startscan()).
 * - When the server is ready to scan (ie. it bursted), we "signal" it
 *   (see scanner_startall()).
 * Note: to simplify, we assume all threads are already waiting when
 *       we are ready to start. Burst times are not that fast... ;-)
 *       It saves a pthread_cond_t.
 */
pthread_mutex_t mutex_init;
pthread_cond_t cond_init;
/*
 */

extern pthread_mutex_t mutex_scanext;
extern pthread_mutex_t mutex_link;

scanthread_s *scanthread = NULL;
scanstats_s *scanstats = NULL;

#define SET_CHECKING_STATE(x, y) x |= SCANTHREAD_CHECKING_##y
#define UNSET_CHECKING_STATE(x, y) x &= ~SCANTHREAD_CHECKING_##y;

long totalscanned = 0;
long totalfound = 0;
long totalcached = 0;

extern scanstats_s *scanstats;
extern time_t startscan_time;
time_t lastnojob = 0;
Boolean syncsent = false;


static void scanner_notify_netcheck();
static void scanner_startscan();


static void
scanner_startscan()
{
   scanthread_s *th;
   
   Debug((DBLBASIC, "scanner_startscan> %lu", (unsigned long) pthread_self()));
   
   scanner_wait_one();
   
   th = get_scanthread(pthread_self());
   
   if (NULL == th) {
      log_error("scanner_startscan: th == NULL !");
      exit(EXIT_FAILURE);
   }
   
   /* Endless loop... */
   for (;;) {
      int i;
      
      pthread_mutex_lock(&mutex_scanext);
      array_fetchitem(th);
      pthread_mutex_unlock(&mutex_scanext); /* Allow others threads to get a next IP */
      
      if (th->counter == 0) {
         
         if (scanstats->waiting_threads >= 1) {
            pthread_mutex_lock(&mutex_init);  /* test&set */
            if (!syncsent) {
               syncsent = true;
               pthread_mutex_unlock(&mutex_init);
               scanner_notify_netcheck();
            } else
               pthread_mutex_unlock(&mutex_init);
         }
         
         /* Wait 2 seconds */
         sleep(2);
         
      } else {
        
         Debug((DBLMAX, "scanner_startscan: OK scan..."));
         
#ifdef SCAN_WINGATE
         SET_CHECKING_STATE(th->state, WINGATE);
         scan_wingate(th);
         UNSET_CHECKING_STATE(th->state, WINGATE);
#endif
#ifdef SCAN_SOCKS
         SET_CHECKING_STATE(th->state, SOCKS);
         scan_socks(th);
         UNSET_CHECKING_STATE(th->state, SOCKS);
#endif
#ifdef SCAN_NEGATIVITY
         SET_CHECKING_STATE(th->state, NEGATIVITY);
         scan_negativity(th);
         UNSET_CHECKING_STATE(th->state, NEGATIVITY);
#endif
#ifdef SCAN_HTTP
         for (i = 0; i < MyPerform->http_ports_count; i++) {
            th->port = MyPerform->http_ports[i];
            SET_CHECKING_STATE(th->state, HTTP);
            scan_proxy(th);
            UNSET_CHECKING_STATE(th->state, HTTP);
         }
#endif
         totalscanned += th->counter;
         
         /* Some hosts might be immediately rechecked if needed, actually
          * because of an EHOSTUNREACH error (the host wasn't reachable thus
          * wasn't scanned).
          */
         
         for (i = 0; i < th->counter; i++) {
            if (th->item[i].dont_cache) {
               pthread_mutex_lock(&mutex_scanext); /* protect the cache */
               log_error("scanner: removing %s from cache (wasn't scanned)",
                   inet_ntoa(th->item[i].addr));
               cache_remove(&th->item[i].addr);
               pthread_mutex_unlock(&mutex_scanext);
            }
         }
      }
   }
}

static void
scanner_notify_netcheck()
{
   time_t now;
   char locbuf[0x80];
   
   time(&now);
   human_time(now - startscan_time, locbuf);
   sendto_console("[%s] Network check completed. (%s)", Me->id, locbuf);

}


/* Execute a scan thread. */
static void*
scanner_run_proc()
{
   scanner_startscan();
   return NULL;
}

/* Launch the server (which will listen for the "CONNECT" replies) */
static void*
scanner_listener_run_proc()
{
   listener_run();
   return NULL;
}


/* Allocate the threads */
void
scanner_init()
{
   int i;
   pthread_t newthread;
   scanthread_s *scanth;
   scanthread_s *scanth_last = NULL;
   
   /* Stats */
   scanstats = (scanstats_s *)xmalloc(sizeof(scanstats_s));
   memset(scanstats, 0, sizeof(scanstats_s));

   scanstats->running_threads = 0;
   scanstats->waiting_threads = MyPerform->scanthreads;
   
   /* Init thread-sync stuffs to start */
   pthread_mutex_init(&mutex_init, NULL);
   pthread_cond_init(&cond_init, NULL);
   
   scanthread = NULL;
   
   listener_init();
   
   /* Create the listener (server) in a new thread */
   if (pthread_create(&newthread, NULL, scanner_listener_run_proc, NULL)) {
      fprintf(stderr, "Server creation failed. Aborting.\n");
      exit(EXIT_FAILURE);
   }
   
   Debug((DBLMAX, "init_scanner> Listener initialized. Initializing scanthreads..."));
   
   printf("Initializing scanning threads.\n");
   
   for (i = 0; i < MyPerform->scanthreads; i++) {
      if (!pthread_create(&newthread, NULL, scanner_run_proc, NULL)) {
         
         Debug((DBLMAX, "scanner_init> th: %lu", (unsigned long) newthread));
         
         scanth = (scanthread_s *) xmalloc(sizeof(scanthread_s));
         memset(scanth, 0, sizeof(scanthread_s));
         
         scanth->threadID = newthread;
         scanth->state = STH_STATE_WAITING;
         scanth->next = NULL;
         
         if (NULL == scanth_last)
            scanthread = scanth;
         else
            scanth_last->next = scanth;

         scanth_last = scanth;
         
      } else {
         fprintf
            (stderr, "\n\nERROR! Not enough system resources to create a new thread "
            "or too much threads; please use a lower value for scanthreads.\n");
         log_error("m_thread:scanner_init: new thread failed (%d/%d)\n",
              i, MyPerform->scanthreads);
         exit(1);
      }
   }
   
   pthread_mutex_unlock(&mutex_init);
}

void
scanner_destroy()
{
   scanthread_s *th, *thn;
   
   for (th = scanthread; th; th = thn) {
      thn = th->next;
      xfree(th);
   }
   
   if (scanstats) {
      xfree(scanstats);
      scanstats = NULL;
   }
   
   pthread_mutex_destroy(&mutex_init);
   pthread_cond_destroy(&cond_init);
   
   Debug((DBLMAX, "scanner_destroy> Done."));
}

void
scanner_wait_one()
{
	/* begin monitor */
	pthread_mutex_lock(&mutex_init);
	
	/* Wait for the condition and unlock the monitor's mutex... */
	pthread_cond_wait(&cond_init, &mutex_init);
	
	/* end monitor */
	pthread_mutex_unlock(&mutex_init);
}


/* We are ready and assume all threads are started and waiting. */
void
scanner_startall()
{   
   /* begin monitor */
   pthread_mutex_lock(&mutex_init);
   
   pthread_cond_broadcast(&cond_init);
   /* We haven't started yet. mutex_init still locked. */
   
   /* end monitor */
   pthread_mutex_unlock(&mutex_init);
   
   /* Ok! Wild animals are liberated. Let the show begin! */
}


/* scanner_killall is mainly for lisuxthreads */
void
scanner_killall()
{
#ifdef HAVE_PTHREAD_KILL
   scanthread_s *th = scanthread;
   
   for(; th; th=th->next)
      pthread_kill(th->threadID, SIGKILL);
#endif
}

int
countscans()
{
   int ret = 0;
   scanthread_s *sth;
   
   for(sth = scanthread; sth; sth = sth->next)
      ret += sth->counter;

   return ret;
}


scanthread_s *
get_scanthread(pthread_t thread)
{
   scanthread_s *ptr;
   
   for(ptr = scanthread; ptr; ptr = ptr->next) {
      if(ptr->threadID == thread)
         return ptr;
   }

   return NULL;                 /* Not found. Argh. */
}


scanthread_s *
get_scanthread_from_ip(
   struct in_addr * ip,
   unsigned int *outport,
   int *outitem)
{
   int i;
   scanthread_s *th = NULL;
   
   pthread_mutex_lock(&mutex_scanext);
   /* for each thread */
   for(th=scanthread; th; th=th->next) {
      if(th->state & STH_STATE_RUNNING) {
         /* for each item */
         for (i = 0; i < th->counter; i++) {
            if ((th->item[i].state & SIT_INPROGRESS) &&
               (memcmp(&th->item[i].addr, ip, sizeof(struct in_addr)) == 0)) {
               if (!(th->item[i].state & SIT_FOUND))
                  th->item[i].state |= SIT_FOUND;
               /* heh, we need only 1 reply! */
               if (th->item[i].state & SIT_RECEIVED) {
                  th = NULL;
               } else {
                  th->item[i].state |= SIT_RECEIVED;
                  
                  if (th->state & SCANTHREAD_CHECKING_HTTP) {
                     if (outport) *outport = th->port;
                     if (outitem) *outitem = i;
                  } else {
                     log_error("get_scanthread_from_ip> found but not in checking_http state");
                     th->item[i].state &= ~SIT_RECEIVED;
                     th->item[i].state &= ~SIT_FOUND;
                     th = NULL;
                  }
               }
               goto exit_loop;
            }
         }
      }
   }
exit_loop:
   pthread_mutex_unlock(&mutex_scanext);
   return th;
}


/* m_scanner.c */



