/* Pxys - An IRC ProxyScan, ms_network.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * Net junction/split handlers.
 *
 * $Id: ms_network.c,v 1.22 2002/12/04 15:50:21 mbuna Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif
#include <unistd.h>

#include "m_array.h"
#include "m_share.h"
#include "m_structs.h"
#include "m_noscan.h"

#include "ms_network.h"
#include "ms_netutils.h"
#include "ms_globals.h"
#include "ms_msgout.h"

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "misc.h"
#include "numnicks.h"
#include "strnlwrcpy.h"
#include "typedefs.h"
#include "yxx.h"

#include "m_log.h"

/* xmalloc at the bottom of include files */
#include "xmalloc.h"

/* Proto's */
static unsigned int get_server_record(const char *server_name, int countID);
static void         set_server_records(const char *server_name,
                                       unsigned int maxusers_count,
                                       unsigned int toasted_count);

static const char *kstats_data_file = "data/stats";

extern pthread_mutex_t mutex_link;
long serv_current = 0;
htab *mHS = NULL;
server_handle_t mSA = NULL;

void
network_init()
{
   /* Create main Hashed Servers table */
   mHS = jenkins_hcreate(2);
   
   /* Pre-allocate FULL servers array. */
   mSA = (server_handle_t) xcalloc(NN_MAX_SERVER, sizeof(void *));
   
   serv_current = 0;
}

void
network_free()
{
   if (mHS) {
      jenkins_hdestroy(mHS);
      mHS = NULL;
   }
   if (mSA) {
      xfree(mSA);
      mSA = NULL;
   }
}

void
server_add(
   unsigned int up_nserv,   /* Uplink numeric */
   char *name,              /* Server name */
   unsigned int nserv,      /* Server numeric ID */
   unsigned int ncap,       /* Server capacity (max clients) */
   int bursting             /* Is Server bursting? */
)
{
   struct Server *newServer;
   unsigned int max_clients = 16;
   
   while (max_clients < ncap)
      max_clients <<= 1;
   --max_clients;
   
   newServer = (struct Server *) xmalloc(sizeof(struct Server));

   /* Server name (we use lowercases in hash table) */
   newServer->name = (char *) xmalloc(strlen(name) + 1);
   strnlwrcpy(newServer->name, name, strlen(name) + 1);
   
   /* Uplink numeric */
   newServer->upx = up_nserv;
   
   /* Server numeric */
   newServer->myx = nserv;
   newServer->nn_mask = max_clients;
   newServer->clients = 0;
   
   newServer->noscan = false;
   newServer->shared = false;
   newServer->bursting = bursting;
   newServer->maxclients = get_server_record(name, 0);
   newServer->currtoasted = 0;
   newServer->lasttoasted = get_server_record(name, 1);
   
   /* Allocate clients numerics array */
   newServer->cliH = (void**) xcalloc(max_clients + 1, sizeof(void*));
   
   pthread_mutex_lock(&mutex_link);
   
   mSA[nserv] = newServer;
   ++serv_current;
   
   hadd(mHS, newServer->name, strlen(newServer->name) + 1, (void*) newServer);
    
   /* Refresh noscan and shared servers list. */
   refresh_noscan_servers();
   refresh_shared_servers();
   
   pthread_mutex_unlock(&mutex_link);
   
   log_error("server_add: server %s[%u] added (capacity: %u)", newServer->name,
	newServer->myx, max_clients + 1);
   Debug((DBLBASIC, "server_add: server %s added (capacity: %u)",
        newServer->name, max_clients + 1));
   
   if (chk_shared(name))
      sendto_console("[%s] Sharing with %s activated.", Me->id, name) ;
}

/* ------------------- NET SPLITS HANDLING ------------------- */


void
do_server_quit(char *name)
{
   if (hfind(mHS, strlwr(name), strlen(name) + 1) == TRUE) {
      
      Debug((DBLMAX, "do_server_quit: server %s found.",
         ((struct Server *)hstuff(mHS))->name));
         
      do_squit((struct Server *)hstuff(mHS));
      
   } else {
      
      Debug((DBLMAX, "current pos: %s",
         ((struct Server *)hstuff(mHS))->name));
      
      Debug((DBLMAX, "do_server_quit: server %s not found.", name));
      log_error("do_server_quit: server %s not found.\n", name);
   
   }
}

void
do_squit(sptr)
struct Server *sptr;
{
   pthread_mutex_lock(&mutex_link);

   if (hfind(mHS, sptr->name, strlen(sptr->name) + 1)) {
   
      do_recursive_squit(sptr);
      
      refresh_shared_servers();   /* Update sharing toggles */
      refresh_relay_services();
   }
#ifdef DEBUG
   else
      Debug((DBLMAX, "do_squit: server %s not found in critical section!", sptr->name));
#endif
   
   pthread_mutex_unlock(&mutex_link);
}

void
do_recursive_squit(sptr)
struct Server *sptr;
{
   int rs = 0;        /* rescan server hash table flag */
   
   Debug((DBLMAX, "do_recursive_squit> %s", sptr->name));
   
   hdel(mHS);
   
   if(chk_shared(sptr->name))
      sendto_console("[%s] Sharing with %s disabled.", Me->id, sptr->name) ;
   
   /* Delete all the users from this server. */
   array_purgeserv(sptr->myx);
   
   /* Process recursively to purge downlinks. */
   if (hfirst(mHS)) do {
      rs = 0;
      if(((struct Server *)hstuff(mHS))->upx == sptr->myx) {
         do_recursive_squit((struct Server *)hstuff(mHS));
         /* restart loop -- cheap way to avoid collision or double delete */
         if(!hfirst(mHS))
            break;
         rs = 1; /* reloop */
      }
   } while (rs || hnext(mHS)); /* reloop from start or try next */
   
   /* Really delete now this server. */
   mSA[sptr->myx] = NULL;
   --serv_current;
   
   set_server_records(sptr->name, sptr->maxclients, sptr->currtoasted + sptr->lasttoasted);
   
   xfree(sptr->name);
   Debug((DBLBASIC, "do_recursive_squit> xfree -> %p", sptr->cliH));
   xfree((void*)sptr->cliH);
   Debug((DBLBASIC, "do_recursive_squit> Server data (%u) free'd.", sptr->myx));
   xfree(sptr);
}

void
refresh_relay_services()
{
   struct Relay_s *relay;
   
   for (relay = MyGline->relay; relay; relay = relay->next) {
      if(!hfind(mHS, relay->serv, strlen(relay->serv) + 1)) {
         /* relay cannot be found -- set it as "deauthed" */
         relay->isauth = 0;
      }
   }
}

/* Check if a server is currently sending its clients/channels to *US*
 */
int is_bursting(struct Server *sptr)
{
   struct Server *uplink;
   int bursting = sptr->bursting;
   int cnt = serv_current;
   
   for (uplink = mSA[sptr->upx]; !bursting && uplink && cnt--; uplink = mSA[uplink->upx])
      bursting = uplink->bursting;
   
   if ((cnt < 0) && mSA[uplink->upx])
      log_error("is_bursting: desync: server (%d) %s (up:%d)\n",
            uplink->myx, uplink->name, uplink->upx);
   return bursting;
}

/* countID = 0 -> maxusers record
 * countID = 1 -> toasted clients record
 */
static unsigned int
get_server_record(
   const char *server_name,
   int countID)
{
   unsigned int count = 0;
   char buffer[256];
   char *tok = NULL, *p = NULL;
   FILE *fp;
   
   fp = fopen(kstats_data_file, "r");
   if (fp != NULL) {
      while (fgets(buffer, sizeof(buffer), fp)) {
         stripline(buffer);
         tok = buffer;
         if ((p = strchr(buffer, ' '))) {
            *p++ = '\0';
            if (*tok && !stricmp(tok, server_name)) {
               tok = p;
               if ((p = strchr(p, ' ')) != NULL) {
                   *p++ = '\0';
                   if (countID == 0)
                      count = (unsigned int) atoi(tok);
                   else
                      count = (unsigned int) atoi(p);
               }
            }
         }
      }
      fclose(fp);
   }
   
   return count;
}

static void
set_server_records(
   const char *server_name,
   unsigned int maxusers_count,
   unsigned int toasted_count)
{
   static const char *ktmpfile = "data/.tmpfile";
   char buffer[256];
   char *tok = NULL, *p = NULL;
   FILE *fp, *tmpfp;
   
   unlink(ktmpfile); /* just to be sure */
   
   tmpfp = fopen(ktmpfile, "w");
   if (tmpfp != NULL) {
      fp = fopen(kstats_data_file, "r");
      if (fp == NULL) {
         fprintf(tmpfp, "%s %u %u\n", server_name, maxusers_count, toasted_count);
      } else {
         while (fgets(buffer, sizeof(buffer), fp)) {
            stripline(buffer);
            p = buffer;
            tok = p;
            if ((p = strchr(p, ' ')))
               *p++ = '\0';
            if (tok && stricmp(tok, server_name))
               fprintf(tmpfp, "%s %s\n", tok, p);
         }
         fprintf(tmpfp, "%s %u %u\n", server_name, maxusers_count, toasted_count);
         fclose(fp);
      }
      fclose(tmpfp);
      unlink(kstats_data_file);
      rename(ktmpfile, kstats_data_file);
   }
}


void
set_servers_records()
{
   if (hfirst(mHS)) do {
      struct Server *sptr = (struct Server *)hstuff(mHS);
      set_server_records(sptr->name, sptr->maxclients,
                         sptr->currtoasted + sptr->lasttoasted);
   } while (hnext(mHS));
}

/* ms_network.c */
